""" Script for generating statistics from sensitive workload data.

It would produce aggregated data per group (privatized). And finally per day 
and hour statistics.

The idea of this script is to create a release of aggregated statistics 
which does not allow for unprivatizing of the gid (which for me is SLURM's
account).

This script will produce:

    (anonymized)
    1) % of jobs submitted per group 
    
    per group
    
    2) hour of the day histogram
    3) day of the week histogram
    4) runtime histogram
    5) allocated cpus histogram

Dependencies:
    
    Python 3.9.5
    pyswf (https://earth.bsc.es/gitlab/mgimenez/pyswf/-/tags/wrapper-impact-paper)
    os (default on 3.9.5)
    argparse (default on 3.9.5)
    numpy (2.0.1)
    pandas (2.2.2)
    opendp (0.10.0)
    

Copyright (C) 2024
Manuel G. Marciani
BSC-CNS Earth Sciences Department
"""

# library to handle differential privacy
# in-house library for handling workload format
import pyswf
import os 
import sys
import argparse
import numpy as np
import pandas as pd
import opendp.prelude as dp
# for timedelta calculations for each one of the timestamps 
from datetime import timedelta
dp.enable_features("contrib", "floating-point")

def noisy_histogram(categories: list, d_in: int, scale: float, data_type = str, bounds = None) -> dp.Transformation:
    """ Creates a transformation for counting categorical data 
    :categories: categories of the data
    :d_in: symmetric difference of the input metric (??) Opendp parameter
    :scale: float that manages the difference between distributions
    :returns: anonymized transformation
    """

    input_space = dp.vector_domain(dp.atom_domain(T=data_type)), dp.symmetric_distance() 

    # assuming symmetric difference for the input dataset distance. which one is better(?)
    histogram = (
        input_space >>
        #dp.t.then_clamp(bounds) >>
        dp.t.then_count_by_categories(categories=categories) 
    )

    return dp.binary_search_chain(
        lambda s: histogram >> dp.m.then_laplace(scale=s),
        d_in=d_in, d_out=scale)  

def variant_statistics(swftrace: pyswf.SWFtrace, d_in: int, budget: float, output_dir: str) -> None:
    """ Create the anonymized statistics and output them both in graphs and the raw data

    :swftrace: input standard workflow format object
    :d_in: symmetric difference of the input metric (??) Opendp parameter
    :budget: output distance assuming it is laplace :) Opendp parameter
    :output_dir: root folder of the output
    """    

    # invariants (not privatized released): submit, start, end, uid, gid (account)
    # we are computing (in timestamp) each of the instants, for then joining per day of the week 
    swftrace.job_list["submit_ts"] = swftrace.job_list.submit_time.apply(lambda instant: timedelta(seconds=instant)) + swftrace.start_time
    swftrace.job_list["start_ts"] = swftrace.job_list.start_time.apply(lambda instant: timedelta(seconds=instant)) + swftrace.start_time
    swftrace.job_list["end_ts"] = swftrace.job_list.end_time.apply(lambda instant: timedelta(seconds=instant)) + swftrace.start_time

    # submit info
    # weekday will be 0, Monday, 1, Tuesday, 2 Wednesday, ...
    swftrace.job_list["submit_day_of_the_week"] = swftrace.job_list["submit_ts"].apply(lambda date: date.weekday())
    swftrace.job_list["submit_hour_of_the_day"] = swftrace.job_list["submit_ts"].apply(lambda date: date.hour)

    # 0 for monday, and so on so forth
    days_in_week = [0,1,2,3,4,5,6]
    hours_in_day = [hour for hour in range(0,24)]

    # this are ad-hoc defined
    division_runtime = [-1, 0, 1, 10, 100, 1000, 10000, 100000, 1e9]
    # -2 so we also grab the badly logged jobs with a -1
    division_alloc_cpus = [-2, 1, 64, 128, 256, 1024, 2048, swftrace.max_procs]

    # calculate the bins according to the edges on top
    swftrace.job_list["runtime_category"] = pd.cut(swftrace.job_list.runtime, division_runtime, labels=division_runtime[:-1])
    swftrace.job_list["alloc_cpus_category"] = pd.cut(swftrace.job_list.alloc_cpus, division_alloc_cpus, labels=division_alloc_cpus[:-1])

    # fill NaNs with -1's 
    swftrace.job_list["runtime_category"] = np.where(swftrace.job_list["runtime_category"].isnull(), -1, swftrace.job_list["runtime_category"])
    swftrace.job_list["alloc_cpus_category"] = np.where(swftrace.job_list["alloc_cpus_category"].isnull(), -2, swftrace.job_list["alloc_cpus_category"])

    #private: per group (slurm account) % of jobs from account (sum
    gids = swftrace.job_list.gid.unique()

    # transformation to anonymized release 
    noisy_per_gid_num_jobs_histogram = noisy_histogram(gids, d_in, budget, dp.u32)

    # output number of jobs per group id
    with open(f"{output_dir}/release/variant/per_gid_num_jobs.csv", "w") as output_per_gid_num_jobs:
        output_per_gid_num_jobs.write("gid,num_jobs\n")
        for key, value in zip(gids,noisy_per_gid_num_jobs_histogram(swftrace.job_list.gid.tolist())):
            output_per_gid_num_jobs.write(f"{key},{value}\n")        

    # create the transformations for each histogram 
    noisy_per_gid_runtime_hist = noisy_histogram(division_runtime[:-1], d_in, budget, dp.i32)
    noisy_per_gid_alloc_cpus_hist = noisy_histogram(division_alloc_cpus[:-1], d_in, budget, dp.i32)
    noisy_per_gid_week_day = noisy_histogram(days_in_week, d_in, budget, dp.i32)
    noisy_per_gid_hour_day = noisy_histogram(hours_in_day, d_in, budget, dp.i32)

    for gid in gids:
        # sensitive group's info 
        gid_jobs = swftrace.job_list[swftrace.job_list.gid == gid]

        with open(f"{output_dir}/release/variant/{gid}_runtime.csv", "w") as gid_runtime_output:
            gid_runtime_output.write("runtime,num_jobs\n")
            # we remove the last right edge of the bin
            for key, value in zip(division_runtime[:-1], noisy_per_gid_runtime_hist(gid_jobs.runtime_category.tolist())):
                gid_runtime_output.write(f"{key},{round(value)}\n")

        with open(f"{output_dir}/release/variant/{gid}_alloc_cpus.csv", "w") as gid_alloc_cpus_output:
            gid_alloc_cpus_output.write("alloc_cpus,num_jobs\n")
            # we remove the last right edge of the bin
            for key, value in zip(division_alloc_cpus[:-1], noisy_per_gid_alloc_cpus_hist(gid_jobs.alloc_cpus_category.tolist())):
                gid_alloc_cpus_output.write(f"{key},{round(value)}\n")                

        with open(f"{output_dir}/release/variant/{gid}_week_day.csv", "w") as gid_week_day:
            gid_week_day.write("day_week,num_jobs\n")
            for key, value in zip(days_in_week, noisy_per_gid_week_day(gid_jobs.submit_day_of_the_week.tolist())):
                gid_week_day.write(f"{key},{round(value)}\n")

        with open(f"{output_dir}/release/variant/{gid}_hour_day.csv", "w") as gid_hour_day:
            gid_hour_day.write("hour_day,num_jobs\n")
            for key, value in zip(hours_in_day, noisy_per_gid_hour_day(gid_jobs.submit_hour_of_the_day.tolist())):
                gid_hour_day.write(f"{key},{round(value)}\n")
                
def arg_parse() -> argparse.ArgumentParser:
    """
    Parse arguments from the command line and check correctness.
    Folder structure will be created if not present 
    :returns: argparse object with the parameters
    """
    parser = argparse.ArgumentParser(description="Anonymize using Differential Privacy workloads which are in Standard Workflow Format")
    parser.add_argument("swf_path", type=str, help="relative or absolute path to the swf file")
    parser.add_argument("out_dir", type=str, help="relative or absolute path to the output directory")
    parser.add_argument("d_in", type=int, help="OpenDP parameter of difference for the symmetric distance")
    parser.add_argument("budget", type=float, help="OpenDP parameter of the laplacian difference between histograms")

    arguments = parser.parse_args()

    arguments.swf_path = os.path.abspath(arguments.swf_path)
    if not os.path.exists(arguments.swf_path):
        sys.exit(f'opendp_release.py: file {arguments.swf_path} does not exist')

    arguments.out_dir = os.path.abspath(arguments.out_dir)
    if not os.path.exists(f"{arguments.out_dir}/release"):
        print(f'opendp_release.py: output folder {arguments.out_dir} does not exist. Creating it...')
        os.mkdir(f"{arguments.out_dir}/release") 
        os.mkdir(f"{arguments.out_dir}/release/invariant") 
        os.mkdir(f"{arguments.out_dir}/release/variant") 

    return arguments

if __name__ == "__main__":
    args = arg_parse()    

    # load data
    workload = pyswf.SWFtrace.from_swf(args.swf_path)
    variant_statistics(workload, args.d_in, args.budget, args.out_dir)
   
